#! /usr/local/bin/python

'''Module---pymake.py

This module creates a skeleton Makefile from a Setup file.

Data Attributes:
    use_configure
    a boolean variable which indicates whether to use the PREDEFS
    values as possibly set by configure or to use the pymake
    built-in values. If you set use_configure to a TRUE value, you
    need to have your Python build tree available so you can
    substitute the appropriate values in PREDEFS. These values may be
    found in config.status.

    PREDEFS_L
    the ordered list for determining the order in which to output the
    predefined Makefile macros.

    PREDEFS
    a dictionary of predefined Makefile macros and their values.

    SYSPREDEFS
    when use_configure is FALSE, SYSPREDEFS is a list of macros
    for overriding the values in PREDEFS.


Functions:
    makefile()

Errors Raised:
    ValueError
'''

### no implied warranty. use at your own risk.
### written by walrus@ans.net


import sys
import regex
import regsub
import string


__version = '$Revision: 1.8 $'

# Possible TODO: integrate pymake into the Python build process
use_configure = 0


# Common Makefile macros.
# A list which determines the order they appear in the output
PREDEFS_L = ['prefix', 'exec_prefix', 'MACHDEP',
	     'CC', 'CCSHARED', 'OPT', 'INCLDIR',
	     'CFLAGS', 'LDSHARED', 'DESTSHARED', 'SO']

# The dictionary which determines their values.
# ---based on a suggestion from anthony baxter
if use_configure:
    PREDEFS={'prefix': '@prefix@',
	     'exec_prefix': '@exec_prefix@',
	     'MACHDEP': '@MACHDEP@',
	     'CC': '@CC@',
	     'CCSHARED': '@CCSHARED@',
	     'OPT': '@OPT@',
	     'INCLDIR': '/usr/local/include', # overridden in makefile()
	     'CFLAGS': '$(OPT) -I$(INCLDIR)',
	     'LDSHARED': '@LDSHARED@',
	     'DESTSHARED': '$(exec_prefix)/lib/python/$(MACHDEP)',
	     'SO': '@SO@'}
else:
    # Most of these definitions may be overridden by system-specific
    # defines in SYSPREDEFS. Others are automatically overridden in
    # makefile().
    PREDEFS={'prefix': '/usr/local/',	# overridden in makefile()
	     'exec_prefix': '$(prefix)',
	     'MACHDEP': 'unknown',	# overridden in makefile()
	     'CC': 'gcc',
	     'CCSHARED': '',
	     'OPT': '-O',
	     'INCLDIR': '/usr/local/include/Py', # overridden in makefile()
	     'CFLAGS': '$(OPT) -I$(INCLDIR)',
	     'LDSHARED': 'ld',
	     'DESTSHARED': '$(exec_prefix)/lib/python/$(MACHDEP)',
	     'SO': '.so'}

    # These are used to override the relevant entries in PREDEFS
    SYSPREDEFS={
	'sunos5': {'LDSHARED': 'ld -G'},
	'next3':  {'CC': 'cc',
		   'LDSHARED': 'ld -r',
		   'SO': '.o'}}


# This bit ensures someone (such as myself) does not add a value to
# PREDEFS, then forgets to make the corresponding addition to
# PREDEFS_L.
pk, pl = PREDEFS.keys(), PREDEFS_L[:]
pk.sort()
pl.sort()
if len(pk) <> len(pl):
    raise ValueError, 'internal PREDEFS/PREDEFS_L length mismatch'
for i in range(len(pk)):
    if pk[i] <> pl[i]:
	raise ValueError, 'internal PREDEFS/PREDEFS_L values mismatch'
del pk, pl


# The function which does all the work
def makefile(setup, mfile = ''):
    '''Generate a Makefile and write it to stdout.

    Takes SETUP string as the name of the Setup file from which to
    generate the Makefile. Optionally takes MFILE string as the name
    of the output file; otherwise output is written to stdout.
    
    Returns 0 on success; 1 on failure.

    The macros in the Pre-definitions section may be overridden in the
    the Setup file. pymake.PREDEFS.keys() contains the macro names one
    should use when overriding.'''

    # Override predefines with system specific predefines
    # --only if not using configure
    if not use_configure:
	for p in SYSPREDEFS[sys.platform].keys():
	    PREDEFS[p] = SYSPREDEFS[sys.platform][p]

    # Find where python is installed and use that to further
    # refine the predefines
    pylib = sys.path[1]
    pytop = regsub.sub('/lib/python', '', pylib)
    pyinc = pytop + '/include/Py'

    if not use_configure:
	PREDEFS['prefix'] = pytop
	PREDEFS['MACHDEP'] = sys.platform

    # XXX Use the installed location for header files
    # XXX This is a problem when trying to build using the source tree
    # XXX header files.
    PREDEFS['INCLDIR'] = pyinc

    try:
	s = open(setup, "r")
    except IOError, msg:
	sys.stderr.write("%s: can't open (%s)\n" % (setup, msg[1]))
	return 1

    if mfile:
	try:
	    m = open(mfile, "w")
	except IOError, msg:
	    sys.stderr.write("%s: can't open (%s)\n" % (mfile, msg[1]))
	    return 1
    else:
	m = sys.stdout

    # These lists are maintained across lines of the Setup file
    DEFS, MODS, RULES = [], [], []

    # Loop over the lines of the Setup file
    while 1:
	line = s.readline()
	if not line:
	    break
	if line[0] == '#':
	    continue
	if 0 < regex.match('.+=.*', line):
	    macro = regsub.sub('\([^=]+\)=[ \t]+.*\n', '\\1', line)
	    defn = regsub.sub('[^=]+=[ \t]+\(.*\)\n', '\\1', line)
	    if PREDEFS.has_key(macro):
		PREDEFS[macro] = defn
	    else:
		DEFS.append(line)
	    continue
	args = string.split(line)

	# These strings and lists are emptied for each Setup line
	cpps, libs, mods, objs, srcs = '', '', [], '', []

	# These regexps were found by examining makesetup
	for idx in range(0, len(args)):
	    arg = args[idx]
	    if arg[0] == '#':
		break
	    if 0 < regex.match('-[IDUC].*', arg):
		cpps = cpps + ' ' + arg
		continue
	    if 0 < regex.match('\(-[A-Zl].*\)\|\(.*\.a\)', arg):
		libs = libs + ' ' + arg
		continue
	    if 0 < regex.match('.*\.\([cC][cC]?\|\(c\+\+\)\)$', arg):
		srcs.append(arg)
		continue
	    if 0 < regex.match('.*\.o$', arg):
		src = regsub.sub('\.o$', '.c', arg)
		srcs.append(src)
		continue
	    if 0 < regex.match('\$.*', arg):
		cpps = cpps + ' ' + arg
		libs = libs + ' ' + arg
		continue
	    if 0 < regex.match('.*\..*', arg):
		s.close()
		raise ValueError, 'bad word ' + arg + ' in ' + line
	    if 0 < regex.match('[a-zA-Z_].*', arg):
		mods.append(arg)
		continue
	    s.close()
	    raise ValueError, 'bad word ' + arg + ' in ' + line

	# Generate the rules for building
	for src in srcs:
	    if 0 < regex.match('.*\.c$', src):
		CC = ' $(CC) '
	    else:
		CC = ' $(CCC) '
	    obj = regsub.sub('\.\(o\|[cC][cC]?\|\(c\+\+\)\)$', '.o', src)
	    RULES.append(obj + ': ' + src + ';'
			 + CC + '$(CCSHARED) $(CFLAGS)' + cpps + ' -c ' + src)
	    objs = objs + ' ' + obj

	# Generate the rules for creating a dynamically loaded module
	#
	# The conditional writing of `mod' does not work well with
	# NeXT modules. Since Python wants the shared object to end
	# in .o, I prefer to use mod'module'.[co] for the source and
	# non-shared object and put the shared object into mod.o.
	#
	# To maintain some consistency with makesetup, however, I
	# ignore my preference and add 'module' to name of the shared
	# object.
	for mod in mods:
	    if 0 < regex.match('.*' + mod + '\.o.*$', objs):
		mod = mod + '$(SO)'
	    else:
		mod = mod + 'module$(SO)'
	    RULES.append(mod + ': ' + objs
			 + '; $(LDSHARED)' + objs + libs + ' -o ' + mod)		
	    MODS.append(mod)

    s.close()

    # Pre-definitions
    m.write('# Pre-definitions\n')
    for d in PREDEFS_L:
	m.write(d + '= ' + PREDEFS[d] + '\n')

    # Definitions
    m.write('\n# Module-supplied Definitions\n')
    for d in DEFS:
	m.write(d)

    # Modules
    m.write('\n# Modules\n')
    m.write('SHAREDMODS=')
    for mod in MODS:
	m.write(' ' + mod)
    m.write('\n')

    # Rules
    #
    m.write('\n# Standard Rules\n')

    m.write('# NB: there is no "install" rule because\n')
    m.write('#     Modules/Makefile does not have one\n')

    m.write('all: $(SHAREDMODS)\n\n')
    m.write('\nsharedinstall: $(DESTSHARED) $(SHAREDMODS)\n')
    m.write('\t\t-for i in dummy $(SHAREDMODS); do \\\n')
    m.write('\t\t    if test -f $$i; then mv $$i $(DESTSHARED)/$$i; fi; done\n')
    m.write('\n$(DESTSHARED):\n')
    m.write('\t\tmkdir $(DESTSHARED)\n')
    m.write('\nclean:\n')
    m.write('\t\t-rm -f *.o core *~ [@,#]* *.old *.orig *.rej\n')

    m.write('\n# Automatically Generated Rules\n')
    for rule in RULES:
	m.write(rule + '\n')
    if mfile:
	m.close()
    return 0


def main():
    n = len(sys.argv)
    if n > 1 and sys.argv[1][0] == '-':
	n = 4
    if n == 3:
	code = makefile(sys.argv[1], sys.argv[2])
    elif n == 2:
	code = makefile(sys.argv[1])
    elif n == 1:
	code = makefile('Setup')
    else:
	sys.stderr.write('Usage: ' + sys.argv[0] + ' [<Setup> [<Makefile>]]\n')
	sys.stderr.write("\tSetup defaults to `Setup'\n")
	sys.stderr.write('\tMakefile defaults to stdout\n')
	code = 1
    sys.exit(code)


if __name__ == '__main__':
    main()
